import inspect

from flasky import errors


class FPipelineContext(object):

    def __init__(self):
        pass

    def create(self, f_list):
        return FPipeline(f_list, self.di, parent_ctx=self)



class FPipeline(object):

    def __init__(self, f_list, di):
        self.f_list = f_list
        self.di = di

    async def run(self, data, **kwargs):

        for f in self.f_list:
            func_args = list(inspect.signature(f).parameters.keys())
            parameters = self._resolve_parameters(func_args)

    async def _resolve_parameters(self, func_args):
        for func_arg in func_args:
            try:
                parameter = await self.di.get(func_arg)
            except errors.ConfigurationError:
                raise errors.FError(
                    err_msg='Function argument is not found in DI '
                )






import copy
import os

import logging

from flasky import events
from flasky.handler import RequestContext
from quark_utilities import query_helpers

from quark_utilities import errors
from quark_utilities import responser
from quark_utilities import temporal_helpers


class GenericResource(object):

    def __init__(self, resource_name, api_prefix=None, collection_name=None,
                 event_name_prefix=None, create_json_schema=None,
                 update_json_schema=None, allowed_ops=None, before_create=None,
                 after_create=None, before_update=None, after_update=None,
                 before_delete=None, after_delete=None):

        self.api_prefix = api_prefix
        self.resource_name = resource_name
        self.event_name_prefix = event_name_prefix
        self.collection_name = collection_name
        self.allowed_ops = allowed_ops or ['CREATE', 'UPDATE', 'READ', 'DELETE', 'QUERY']

        self.logger = logging.getLogger('resource.' + self.resource_name + '.logger')

        #: Json Schemas
        self._create_json_schema = create_json_schema
        self._update_json_schema = update_json_schema

        #: Hooks
        self._before_create = before_create or []
        self._after_create = after_create or []
        self._before_update = before_update or []
        self._after_update = after_update or []
        self._after_delete = after_delete or []
        self._before_delete = before_delete or []

    def before_create(self, f):
        self._before_create.append(f)
        return f

    def resolve_url(self, url=None):
        resolved = os.path.join('/', self.api_prefix or '', self.resource_name, url or '')

        if resolved.endswith('/'):
            return resolved[:-1]

        return resolved

    @property
    def repository_name(self):
        return self.collection_name + '_repository'

    async def repository(self):
        return await RequestContext.current_app().di.get(self.repository_name)

    def resolve_event_name(self, event_type):
        if event_type == 'create':
            return self.event_name_prefix + 'CreatedEvent'
        elif event_type == 'update':
            return self.event_name_prefix + 'UpdatedEvent'
        elif event_type == 'delete':
            return self.event_name_prefix + 'DeletedEvent'
        else:
            raise errors.QError(
                err_msg='Unsupported event type<{}>'.format(event_type),
                err_code='errors.internalError'
            )

    async def __run_function_pool(self,handler, func_pool, **kwargs):
        document = kwargs.get('document', None)
        if not document:
            raise errors.FError(
                err_msg='Internal error occured...',
                err_code='errors.internalError',
                status_code=500
            )

        for func in func_pool:
            document = await handler.di.inject_and_run(func, **kwargs)

        return document

    async def create(self, handler, *args, **kwargs):
        domain_id = args[0]
        subject = RequestContext.current_data()['subject']

        resource = handler.body_as_json()

        resource.update({
            'domain_id': domain_id,
            'membership_id': subject.membership_id,
            'sys': {
                'created_at': temporal_helpers.utc_now(),
                'created_by': subject.token['prn'],
                'cid': (await handler.di.cid_generator.generate(self.collection_name)),
                'version': 1
            }
        })

        repository = await handler.di.get(self.repository_name)

        resource = await self.__run_function_pool(
            handler,
            self._before_create,
            resource=resource
        )

        saved = await repository.save(resource)

        resource = await self.__run_function_pool(
            handler,
            self._after_create,
            resource=resource
        )

        await events.dispatch(events.Event(
            type=self.resolve_event_name('create'),
            document=resource,
            domain_id=domain_id,
            token=subject.token,
            sys={
                'created_at': temporal_helpers.utc_now(),
                'created_by': 'system'
            }
        ))

        responser.to_response(
            handler,
            201,
            documents=saved
        )

    async def query(self, handler, *args, **kwargs):
        domain_id = args[0]

        where, select, skip, limit , sort = query_helpers.parse(handler)

        repository = await handler.di.get(self.repository_name)

        where.update({
            'domain_id': domain_id
        })

        contents, total_count = await repository.query(
            where=where,
            select=select,
            skip=skip,
            limit=limit,
            sort=sort
        )

        responser.to_response(
            handler,
            200,
            documents=contents,
            count=total_count
        )


    async def get(self, handler, *args, **kwargs):
        domain_id, content_id = args[0], args[1]

        repository = await self.repository()

        resource = await repository.find_one_by(where={
            '_id': content_id,
            'domain_id': domain_id
        })

        responser.to_response(
            handler,
            200,
            documents=resource
        )

    async def update(self, handler, *args, **kwargs):
        domain_id, resource_id, subject = args[0], args[1], RequestContext.get_param('subject')

        repository = await self.repository()

        resource = await repository.find_one_by(
          where={
              '_id': resource_id,
              'domain_id': domain_id
          }
        )

        _resource = copy.deepcopy(resource)

        extra_data = handler.body_as_json()
        extra_data.pop('sys', None)
        extra_data.pop('membership_id', None)
        extra_data.pop('domain_id', None)
        extra_data.pop('_id', None)

        resource.update(extra_data)
        resource['sys'].update({
            'modified_at': temporal_helpers.utc_now(),
            'modified_by': subject.token['prn'],
            'version': resource['sys'].get('version', 1) + 1
        })

        resource = self.__run_function_pool(
            handler,
            self._before_update,
            resource=resource,
            _resource=_resource
        )

        await repository.replace(resource)

        resource = await self.__run_function_pool(
            handler,
            self._after_update,
            resource=resource,
            _resource=_resource
        )

        await events.dispatch(
            events.Event(
                type=self.resolve_event_name('update'),
                prior=_resource,
                document=resource,
                user=subject.token,
                domain_id=domain_id
            )
        )

        responser.to_response(
            handler,
            200,
            documents=resource
        )

    async def delete(self, handler, *args, **kwargs):
        domain_id, resource_id = args[0], args[1]
        subject = RequestContext.get_param('subject')

        repository = await handler.di.get(self.repository_name)

        document_query = {
            '_id': resource_id,
            'domain_id': domain_id
        }

        resource = await repository.find_one_by(where=document_query)

        resource = await self.__run_function_pool(
            handler,
            self._before_delete,
            resource=resource
        )

        await repository.remove_one_by(
            where={
                '_id': resource_id,
                'domain_id': domain_id
            }
        )

        resource = await self.__run_function_pool(
            handler,
            self._after_delete,
            resource=resource
        )

        await events.dispatch(
            events.Event(
                type=self.resolve_event_name('delete'),
                document=resource,
                user=subject.token
            )
        )

        responser.to_response(
            handler,
            204
        )

    def register(self, app):
        self.logger.info("Resource initialized on <{}> URL".format(self.resolve_url()))

        if 'CREATE' in self.allowed_ops:
            if self._create_json_schema:
                @app.api(
                    endpoint=self.resolve_url(),
                    method='POST',
                    json_schema=self._create_json_schema
                )
                async def create_resource(handler, *args, **kwargs):
                    await self.create(handler, *args, **kwargs)
            else:
                @app.api(
                    endpoint=self.resolve_url(),
                    method='POST'
                )
                async def create_resource(handler, *args, **kwargs):
                    await self.create(handler, *args, **kwargs)

        if 'QUERY' in self.allowed_ops:
            @app.api(
                endpoint=self.resolve_url('_query'),
                method='POST'
            )
            async def query_resource(handler, *args, **kwargs):
                await self.query(handler, *args, **kwargs)

        if 'READ' in self.allowed_ops:
            @app.api(
                endpoint=self.resolve_url('([^/]+)'),
                method='GET'
            )
            async def read_resource(handler, *args, **kwargs):
                await self.get(handler, *args, **kwargs)


        if 'UPDATE'in self.allowed_ops:
            if self._update_json_schema:
                @app.api(
                    endpoint=self.resolve_url('([^/]+)'),
                    method='PUT',
                    json_schema=self._update_json_schema
                )
                async def update_resource(handler, *args, **kwargs):
                    await self.update(handler, *args, **kwargs)
            else:
                @app.api(
                    endpoint=self.resolve_url('([^/]+)'),
                    method='PUT',
                )
                async def update_resource(handler, *args, **kwargs):
                    await self.update(handler, *args, **kwargs)

        if 'DELETE' in self.allowed_ops:
            @app.api(
                endpoint=self.resolve_url('([^/]+)'),
                method='DELETE'
            )
            async def delete_resource(handler, *args, **kwargs):
                await self.delete(handler, *args, **kwargs)





import json
import traceback

import logging
from functools import singledispatch

from quark_utilities import responser
from quark_utilities.json_helpers import bson_to_json

from flasky import FError, FWarning

logger = logging.getLogger('error_handler')

def register_error_handler(app):

    @app.error_handler()
    async def handle_exc(handler, e, method_definition):
        _handle_exc(e, handler, method_definition=method_definition)

@singledispatch
def _handle_exc(e, handler, method_definition, **kwargs):
    logger.exception(msg="Unhandled exception occured....")

    handler.write(json.dumps(
        {
            "error": {
                "code": "errors.internalError",
                "message": str(e),
                "context": {
                    "stacktrace": traceback.format_exc()
                }
            }
        }
    ))
    handler.set_status(500)
    handler.set_header("Content-Type", "application/json")

@_handle_exc.register(FError)
def handle_ferror(e, handler, **kwargs):
    logger.exception(msg='Internal error occured...')

    handler.write(json.dumps(
        {
            "error": {
                "code": e.err_code,
                "message": e.err_msg,
                "context": e.context,
            }
        }, default=bson_to_json))

    handler.set_status(e.status_code)
    handler.set_header("Content-Type", "application/json")

@_handle_exc.register(FWarning)
def handle_fwarning(e, handler, **kwargs):
    logger.exception(msg='Internal error occured...')

    handler.write(json.dumps(
        {
            "warning": {
                "code": e.err_code,
                "message": e.err_msg,
                "context": e.context,
            }
        }, default=bson_to_json))

    handler.set_status(e.status_code)
    handler.set_header("Content-Type", "application/json")

class QError(Exception):

    def __init__(self, err_msg, err_code=None, status_code=500,
                 context=None, reason=None):
        super(QError, self).__init__(err_msg)
        self.err_code = err_code or "errors.internalError"
        self.err_msg = err_msg or "Internal error occured please contact your admin"
        self.status_code = status_code
        self.context = context or {}
        self.reason = reason


def handle_exc(handler, err):
    responser.to_response(
        handler,
        500,
        documents= {
        "error": {
            "code": "errors.internalError",
            "message": str(err),
            "context": {
                "stacktrace": traceback.format_exc()
            }
        }
    })



def handle_q_error(handler, err):
    responser.to_response(
        handler,
        err.status_code,
        documents={
            "error": {
                "code": err.err_code,
                "message": err.err_msg,
                "context": err.context,
            }
        }
    )


def handle_error(handler, err, method_def):
    if type(err) == QError:
        handle_q_error(handler, err)
    else:
        handle_exc(handler, err)

    logger.exception('Error occured')


class QErrorHandler(object):

    def handle(self, handler, err):
       responser.to_response(
           handler,
           err.status_code,
           documents={
                    "error": {
                        "code": err.err_code,
                        "message": err.err_msg,
                        "context": err.context,
                }
            }
       )




import json
import logging

from concurrent.futures import ThreadPoolExecutor
from tornado import gen

from src.helpers.json_helpers import bson_to_json
from src.utils import temporal_helpers
from src.utils.db import GenericRepository

_logger = logging.getLogger("cm.events")

__HANDLERS = {}
__HANDLERS[None] = []

_executor = ThreadPoolExecutor(2)

class Event(object):
    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

class EventPersister(GenericRepository):


    async def on_event(self, event):
        data = {
            "type": event.__class__.__name__
        }

        data.update(event.__dict__)

        data['sys'] = {}
        data['sys']['created_at'] = temporal_helpers.utc_now()

        await self.save(json.loads(json.dumps(data, default=bson_to_json)))

        _logger.info("Persisted event.")

@gen.coroutine
def dispatch(event):
    _e_type = type(event)

    for handler in __HANDLERS[None] or []:
        _logger.info("Handler<{}> runned for event<{}>".format(
            handler.__name__, type(event).__name__
        ))

        yield _executor.submit(handler, event)

    if _e_type not in __HANDLERS:
        return

    for handler in __HANDLERS[_e_type]:
        _logger.info("Handler<{}> runned for event<{}>".format(
            handler.__name__, type(event).__name__
        ))
        yield _executor.submit(handler, event)


def subscribe(event_type, f):
    global __HANDLERS

    if event_type not in __HANDLERS:
        __HANDLERS[event_type] = []

    __HANDLERS[event_type].append(f)

    _logger.info(
        "Function<{}> subscribed event<{}>."
            .format(f.__name__, event_type.__name__)
    )

def subscribe_global(f):
    global __HANDLERS
    __HANDLERS[None].append(f)







from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorClient
from tornado.testing import AsyncTestCase, gen_test
from tornado.platform.asyncio import AsyncIOMainLoop

from quark_utilities.db import GenericRepository, create_id, GRDuplicateKeyError, GRResourceNotFoundError, normalize_ids


class TestGenericRepository(AsyncTestCase):

    def setUp(self):
        super(TestGenericRepository, self).setUp()
        self.client = AsyncIOMotorClient("localhost")
        self.repo = GenericRepository(self.client.test.test_collection)

    def get_new_ioloop(self):
        return AsyncIOMainLoop()

    @gen_test
    async def test_save_method(self):
        repo = GenericRepository(self.client.test.test_collection)
        saved = await repo.save({
            "email": "test@test.com.tr"
        })

        self.assertIsNotNone(saved['_id'])

        founded = await repo.find_one_by_id(saved['_id'])

        self.assertEquals(founded['_id'], saved['_id'])

    @gen_test
    async def test_save_method_should_raise_duplicate_key_error(self):
        _id = create_id()

        await self.repo.save({
            "_id": _id,
            'key': 'val'
        })

        try:
            await self.repo.save({
                '_id': _id,
                'key': 'val'
            })
        except GRDuplicateKeyError as e:
            self.assertEquals(e.err_code, "errors.documentConflict")
            return

        self.assertFalse(True, "exception did not raised")

    @gen_test
    async def test_find_one_by_id_should_raise_exc_if_doc_not_found(self):
        try:
            await self.repo.find_one_by_id(_id="test")
        except GRResourceNotFoundError as e:
            self.assertEquals(e.err_code, "errors.notFoundError")
            return

        self.assertFalse(True, "Exception did not raised...")


    @gen_test
    async def test_query_should_return_all_items(self):
        doc_1 = {
            '_id': create_id(),
            'test': 'val'
        }

        doc_2 = {
            '_id': create_id(),
            'test': 'val2'
        }

        await self.repo._collection.drop()

        await self.repo._collection.save(doc_1)
        await self.repo._collection.save(doc_2)


        docs, count = await self.repo.query()

        self.assertEquals(count, 2)
        self.assertEquals(len(docs), 2)

    @gen_test
    async def test_update_one(self):
        doc = {
            '_id': create_id()
        }

        await self.repo.save(doc)

        doc.update({
            "test": "val"
        })
        await self.repo.replace_one(doc)

        founded = await self.repo.find_one_by_id(doc['_id'])
        self.assertEquals(founded['test'], doc['test'])

    @gen_test
    async def test_delete_one_by(self):
        doc = {
            '_id': create_id()
        }

        await self.repo.save(doc)

        await self.repo.delete_one_by_id(_id=doc['_id'])

        founded = self.repo.find_one_by_id(_id=doc['_id'], raise_exec=False)

        self.assertIsNotNone(founded)

    @gen_test
    async def test_normalize_ids(self):
        where = normalize_ids({
            "_id": {
                '$in': [str(create_id()), str(create_id())]
            }
        })

        for _id in where['_id']['$in']:
            self.assertEquals(type(_id), ObjectId)


from bson import ObjectId
from pymongo.errors import DuplicateKeyError


def maybe_object_id(maybe_id):
    if isinstance(maybe_id, ObjectId):
        return maybe_id
    elif not ObjectId.is_valid(maybe_id):
        return maybe_id
    else:
        return ObjectId(maybe_id)


def normalize_ids(where):
    if '_id' in where:
        if type(where['_id']) == dict and "$in" in where['_id']:
            _ids = [
                maybe_object_id(_id)
                for _id in where['_id']['$in']
            ]
            where['_id']['$in'] = _ids
            return where
        elif type(where['_id']) == str:
            where['_id'] = normalize_ids(where['_id'])
            return where
    return where


def create_id():
    return ObjectId()


class GRError(Exception):
    def __init__(self, err_code=None, message=None, context=None):
        self.err_code = err_code
        self.message = message
        self.context = context


class GRDuplicateKeyError(GRError):

    def __init__(self, collection_name, context):
        super(GRDuplicateKeyError, self).__init__(
            err_code="errors.documentConflict",
            message="There is already existing document in collection<{}> with given key"
                .format(collection_name),
            context=context
        )


class GRResourceNotFoundError(GRError):
    def __init__(self, collection_name, context):
        super(GRResourceNotFoundError, self).__init__(
            err_code="errors.notFoundError",
            message="Resource not found in collection<{}> with given key.".format(collection_name),
            context=context
        )


class GenericRepository(object):

    def __init__(self, collection):
        self._collection = collection
        self._max_document_limit = 500

    @property
    def name(self):
        return self._collection.name

    async def find_one_by_id(self, _id, raise_exec=True):
        return await self.find_one_by(
            where={
                "_id": maybe_object_id(_id)
            },
            raise_exec=raise_exec
        )

    async def find_one_by(self, where=None, select=None, raise_exec=True):
        document = await self._collection.find_one(
            where,
            select
        )

        if not document and raise_exec:
            raise GRResourceNotFoundError(self.name, context={
                "where": where,
                "select": select
            })

        return document

    async def save(self, document):
        if "_id" not in document:
            document['_id'] = create_id()
        try:
            await self._collection.insert_one(document)
        except DuplicateKeyError as e:
            raise GRDuplicateKeyError(
                self.name,
                context={
                    'document': document
                }
            ) from e

        return document

    async def replace_one(self, document, raise_exec=True):
        if not '_id' in document:
            raise GRError(
                err_code="errors.badDocument",
                message="Can't update document without _id field.",
                context={
                    "document": document
                }
            )

        key = {
            "_id": maybe_object_id(document['_id'])
        }

        result = await self._collection.replace_one(key, document)

        if result.modified_count == 0 and raise_exec:
            raise GRResourceNotFoundError(self.name,
                context={
                    "key": key
                }
            )

        return document

    async def delete_one_by_id(self, _id):
        _id = maybe_object_id(_id)

        key = {
            "_id": _id
        }

        result = await self._collection.delete_one(key)

        if result.deleted_count == 0:
            raise GRResourceNotFoundError(
                self.name,
                context={
                    'key': key
                }
            )

    async def query(self, where=None, select=None, skip=None,
                    limit=None, sort=None):

        where = normalize_ids(where) if where else {}

        cur = self._collection.find(
            where,
            select
        )

        total_document_count = await cur.count()

        if limit: cur.limit(int(limit))
        if skip: cur.skip(int(skip))
        if sort: cur.sort(sort)

        documents = await cur.to_list(length=None)

        return documents, total_document_count




from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from smtplib import SMTP


class EmailClientError(Exception):
    pass

def to_ls(addr):
    if isinstance(addr, str):
        addr = [addr]
    elif isinstance(addr, list):
        addr = ','.join(addr)
    return addr

class EmailClient(object):

    def __init__(self, host=None, port=None,
                 username=None, password=None):
        self.host = host
        self.port = port
        self.username = username
        self.password = password

    def send_mail(self, _from=None, receivers=None, data=None, cc=None,
                  bcc=None, subject=None, text=None, html=None):

        msg = MIMEMultipart('alternative')

        if not receivers or not _from:
            raise EmailClientError(
                'Mail can not be sent without _to or _from')

        cc = to_ls(cc)
        bcc = to_ls(bcc)

        if subject: msg['Subject'] = Header(subject, 'utf-8')
        if cc: msg['CC'] = cc
        if bcc: msg['BCC'] = bcc
        if text: msg.attach(MIMEText(text, 'plain', 'utf-8'))
        if html: msg.attach(MIMEText(html, 'html', 'utf-8'))

        msg['From'] = formataddr((str(_from[0]), _from[1]))
        msg['To'] = to_ls(receivers)

        client = SMTP(host=self.host, port=self.port)
        if self.username and self.password:
            client.login(self.username, self.password)
        client.sendmail(
            msg=msg.as_string(),
            to_addrs=receivers,
            from_addr=_from
        )

        client.quit()












import json

from quark_utilities import json_helpers


def to_response(handler, status_code, documents=None, count=None, formatter=None):
    handler.set_status(status_code)
    if not documents:
        return

    if type(documents) == dict:
        handler.write(
            json.dumps(
                formatter(documents) if formatter else documents, default=json_helpers.bson_to_json
            ))
    elif isinstance(documents, (list, set)):
        if formatter:
            documents = [formatter(document) for document in documents]

        handler.write(
            json.dumps({
                "data": {
                    "items": documents,
                    "count": count or len(documents)
                }
            }, default=json_helpers.bson_to_json))

    handler.set_header('Content-Type', 'application/json')


from bson.json_util import loads
import logging


def get_skip(handler):
    skip = handler.get_argument('skip', 0)
    return int(skip)


def get_limit(handler, check_max_limit=True):
    max_limit = 1000
    limit = 0
    limit = handler.get_argument('limit', None)
    if limit:
        if check_max_limit and int(limit) > max_limit:
            limit = max_limit
        return int(limit)


def get_select(handler):
    select = None
    select_param = handler.get_argument('select', None)
    if select_param:
        select = {}
        fields = select_param.split(',')
        for field in fields:
            field = field.strip()
            if field:
                select[field] = 1

    return select


def get_where(handler):
    try:
        json = loads(handler.request.body.decode("utf-8")).get("where", None)
        return json
    except Exception as ex:
        logging.error(ex)
        return None


def get_sort(handler):
    sort = None
    order_by = handler.get_argument('sort', None)
    if order_by:
        sort = []
        clauses = order_by.split(',')
        for clause in clauses:
            clause = clause.strip()
            if clause:
                if ' ' not in clause:
                    clause = clause + ' asc'
                sort.append(
                    (clause.split(' ')[0], -1 if clause.split(' ')[1] == 'desc' else 1))

    return sort

def parse_sort(order_by_str):
    sort = []
    clauses = order_by_str.split(',')
    for clause in clauses:
        clause = clause.strip()
        if clause:
            if ' ' not in clause:
                clause = clause + ' asc'
            sort.append(
                (clause.split(' ')[0], -1 if clause.split(' ')[1] == 'desc' else 1))

    return sort


def parse(handler):
    where = get_where(handler)
    select = get_select(handler)
    limit = get_limit(handler)
    sort = get_sort(handler)
    skip = get_skip(handler)
    return where, select, limit, sort, skip


import json

from tornado import gen

from flasky.handler import RequestContext
from . import errors
from bson import ObjectId
from flasky.errors import FlaskyTornError
from pymongo.errors import DuplicateKeyError

__all__ = ["maybe_object_id", "Repository", "create_id", "GenericRepository"]


class RecordNotExistsError(FlaskyTornError):
    def __init__(self, record_type, record_id):
        super().__init__(status_code=404, err_code="errors.resourceNotFound")
        self.record_type = record_type
        self.record_id = record_id

class IDMismatchError(FlaskyTornError):
    def __init__(self, given_id, data_id):
        super().__init__(message="errors.idMismatchError", err_code="errors.idMismatchError", status_code=400)
        self.given_id = given_id
        self.data_id = data_id

class FlaskyDuplicateKeyError(FlaskyTornError):
    def __init__(self, collection_name, key, pymongo_exc=None):
        super().__init__(status_code=409,
                         message="There is already existing record at <{}> collection"
                                 "with given <{}>key ".format(collection_name, key),
                         err_code="errors.resourceAlreadyExists",
                         reason=pymongo_exc)
        self.collection_name = collection_name
        self.key = key


def raise_resource_not_found_err(collection_name, key):
    raise errors.FError(
        err_code="errors.resourceNotFound",
        status_code=404,
        err_msg="Not found any resource in collection<{}> with given key <{}>".format(collection_name, key),
        context={
            'collection_name': collection_name,
            'key': key
        }
    )


def maybe_object_id(maybe_id):
    if isinstance(maybe_id, ObjectId):
        return maybe_id
    elif not ObjectId.is_valid(maybe_id):
        return maybe_id
    else:
        return ObjectId(maybe_id)

def normalize_ids(where):
    if '_id' in where:
        if type(where['_id']) == dict and "$in" in where['_id']:
            _ids = [
                maybe_object_id(_id)
                for _id in where['_id']['$in']
            ]
            where['_id']['$in'] = _ids
            return where
        elif type(where['_id']) == str:
            where['_id'] = maybe_object_id(where['_id'])
            return where
    return where


def ensure_domain_id(where):
    ctx = RequestContext.current_data()
    if not ctx or 'domain_id' not in ctx:
        if '_id' not in where and '$in' not in where:
            raise errors.FError(
                err_msg='ID-Free or DomainID-Free query detected...',
                err_code='errors.internalError',
                status_code=500,
                context={
                    'query': where
                }
            )
        return where

    if 'domain_id' not in where and ('_id' not in where and '$in' not in where):
        raise errors.FError(
            err_msg="DomainID-free query detected...",
            err_code='errors.internalError',
            status_code=500,
            context={
                'query': where
            }
        )

    if 'domain_id' not in where:
        where['domain_id'] = ctx['domain_id']

    return where

def _pre_process_where(where):
    normalized_where = ensure_domain_id(normalize_ids(where))
    return normalized_where

def create_id():
    return ObjectId()

class Repository(object):

    def __init__(self, collection):
        self.collection = collection

    @property
    def collection_name(self):
        return self.collection.name

    async def query(self, select=None, where=None,
                    sort=None, skip=None, limit=None, return_count=False):
        cur = self.collection.find(where, select)
        count  = await cur.count()

        if skip:
            cur.skip(skip)

        if sort:
            cur.sort(sort)

        if limit:
            cur.limit(limit)

        records = await cur.to_list(length=None)

        if return_count:
            return records, count

        return [record
                for record in records
                if not record.get("is_deleted", False)]

    async def save(self, document, force=False):
        if "_id" not in document:
            document["_id"] = create_id()
        try:
            if force:
                await self.collection.replace_one(
                    {'_id': document['_id']},
                    document,
                    upsert=True
                )
            else:
                await self.collection.save(document)
        except DuplicateKeyError:
            raise errors.QError(
                err_code='errors.duplicateKeyError',
                status_code=409,
                err_msg='There is already existing record in collection<{}>.'.format(self.collection_name),
                context={
                    "collection_name": self.collection_name,
                })

        return document

    async def update(self, _id, data):
        #: Data object should not have id field in it
        #: if it has it should be the same id with given
        #: id
        if "_id" in data:
            if str(_id) != str(data["_id"]):
                raise IDMismatchError(_id, data["_id"])

            data["_id"] = maybe_object_id(data["_id"])

        try:
            await self.collection.replace_one({
                "_id": maybe_object_id(_id)
            }, data)
            return data
        except DuplicateKeyError as e:
            raise FlaskyDuplicateKeyError(
                collection_name=self.collection_name,
                key=data.get("_id"),
                pymongo_exc=e
            )

    async def delete(self, _id):
        record = await self.find(_id=_id)
        if not record:
            raise RecordNotExistsError(self.collection.name, _id)
        await self.collection.remove(record["_id"], record)

    async def delete_one_by(self, where=None):
        record = await self.find_one_by(where=where)
        if not record:
            raise RecordNotExistsError(self.collection.name, where.get("_id", None))

        if "_id" in where:
            where["_id"] = maybe_object_id(where=where)

        await self.collection.deleteOne(where=where)

    async def remove_one_by(self, where=None, raise_exec=True):
        where = normalize_ids(where)

        result = await self.collection.remove(where)

        if result["n"] == 0 and raise_exec:
            raise_resource_not_found_err(
                self.collection,
                where
            )

    async def delete_many_by(self, where=None):
        record = await self.find_one_by(where=where)
        if not record:
            raise RecordNotExistsError(self.collection.name, where.get("_id", None))

        await self.collection.deleteMany(where=where)

    async def find(self, _id=None, raise_exec=True):
        document = await self.collection.find_one({"_id": maybe_object_id(_id)})
        if not document and raise_exec:
            raise RecordNotExistsError(self.collection_name, _id)
        return document

    async def find_one_by(self, where=None, select=None, raise_exec=True):
        if "_id" in where:
            where["_id"] = maybe_object_id(where["_id"])
        document = await self.collection.find_one(where, select)
        if not document and raise_exec:
            raise RecordNotExistsError(self.collection_name, where)
        return document

    async def update_one_by(self, where=None, update=None):
        if '_id' in where:
            where['_id'] = maybe_object_id(where['_id'])

        return await self.collection.update(where, update)




class GenericRepository(object):

    def __init__(self, collection):
        self.collection = collection

    @property
    def collection_name(self):
        return self.collection.name

    async def save(self, document, force=False):
        if "_id" not in document:
            document["_id"] = create_id()
        try:
            if force:
                await self.collection.replace_one(
                    {'_id': document['_id']},
                    document,
                    upsert=True
                )
            else:
                await self.collection.save(document)
        except DuplicateKeyError:
            raise errors.FError(
                err_code="errors.duplicateKeyError",
                status_code=409,
                err_msg="There is already existing record in collection<{}>."
                        .format(self.collection_name),
                context={
                    "collection": self.collection_name}
            )


        return document

    async def replace(self, document):
        if "_id" not in document:
            raise errors.FError(
                err_code="errors.invalidResource",
                status_code=400,
                err_msg="Document is invalid for this operation<{}>"
                    .format('UPDATE'),
                context={
                    "document": document,
                    "collection": self.collection}
            )

        where = {
            '_id': maybe_object_id(document.get('_id'))
        }

        return await self.collection.replace_one(where, normalize_ids(document))

    async def update_one_by(self, where, command):
        where = _pre_process_where(where)

        update_result = await self.collection.update_one(where, command, multi=False)

        if update_result.modified_count == 0:
            raise_resource_not_found_err(self.collection_name, where)



    async def find_one_by(self, where=None, select=None, raise_exec=True):
        where = _pre_process_where(where)

        founded = await self.collection.find_one(where, select)

        if not founded and raise_exec:
            raise_resource_not_found_err(self.collection_name, where)

        return founded

    async def find_one_by_id(self, _id, raise_exec=True):
        founded = await self.find_one_by(where={
            "_id": maybe_object_id(_id),
            'domain_id': RequestContext.get_param('domain_id')
        },
        raise_exec=raise_exec)

        return founded

    async def remove_one_by(self, where=None):
        where = _pre_process_where(where)

        result = await self.collection.remove(where)

        if result["n"] == 0:
            raise_resource_not_found_err(self.collection_name, where)


    async def remove_one_by_id(self, _id):
        await self.remove_one_by(
            where={
                "_id": maybe_object_id(_id)
            }
        )

    async def query(self, where=None, select=None, limit=None,
              skip=None, sort=None):

            where = _pre_process_where(where)

            if not select:
                select = None

            if not limit or limit > 1000:
                limit = 500

            cur = self.collection.find(where, select)

            total_count = await cur.count()

            if skip:
                cur.skip(int(skip))

            if limit:
                cur.limit(int(limit))

            if sort:
                cur.sort(sort)

            items = await cur.to_list(length=None)

            return (items, total_count)









import json

import jwt
import logging
import user_agents
from urllib.parse import urlparse

from flasky.errors import FlaskyTornError, InvalidTokenError

from quark_utilities import errors

logger = logging.getLogger("utilities.security")


class AgentIsNotAuthorized(Exception):

    def __init__(self, status_code=403, message="Agent is not authorized for this action",
                 required_permission=None, required_domain=None,
                 err_code=None, required_membership=None):

        self.status_code = status_code
        self.required_permission = required_permission
        self.required_membership = required_membership
        self.required_domain = required_domain
        self.message = message
        self.err_code = err_code or "errors.notAuthorized"

class Subject(object):

    def __init__(self, token=None, user=None, request=None):
        self.token = token
        self.user = user
        self.request = request

    @property
    def prn(self):
        return self.token.get("prn")

    @property
    def exp(self):
        return self.token.get("exp")

    @property
    def membership_id(self):
        return self._get_data('membership_id')

    def log(self):
        return {
            'ip': self._get_ip(),
            'useragent': self._parse_user_agent(),
            'referer': self._get_referer()
        }

    def _get_data(self, key):
        return self.user.get(key, None) \
            if self.user \
            else self.token.get(key, None)

    def has_permission(self, required_permission):
        return implies_any(
            self.token["roles"], required_permission)

    def has_domain(self, domain_id):
        return domain_id in [domain["_id"] for domain in self.token.get("domains", [])]

    def _get_domain(self, domain_id):
        for domain in self.token["domains"]:
            if domain["domain_id"] == domain_id:
                return domain

    @property
    def is_anonymous(self):
        return True if not self.token else False

    def has_permission_in_root(self, permission):
        return implies_any(self.token["permissions"], permission)

    def has_permission_in_domain(self, domain_id, permission):
        if not self.token:
            return False

        if implies_any(self.token["permissions"], permission):
            return

        domain = self._get_domain(domain_id)
        if not domain or not implies_any(domain.get("permissions", []), permission):
            return False

        return True


    def ensure_permission_in_domain(self, domain_id, permission):
        if not self.is_anonymous and self.has_permission_in_root(permission):
            for domain in self.token["domains"]:
                if str(domain_id) == (domain.get("_id") or domain.get('domain_id')):
                    return True

        elif not self.is_anonymous and self.has_permission_in_domain(domain_id, permission):
            return True

        raise errors.QError(
            err_msg="Subject is not authorized for this action<{}>".format(permission),
            err_code="errors.notAuthorized",
            context={
                "token": self.token
            }
        )

    def has_application(self, app_id):
        return app_id in [app["_id"] for app in self.token.get("applications", [])]

    def _get_referer(self):
        try:
            parsed = urlparse(self.request.headers.get('Referer'))
            return {
                'netloc': parsed.netloc.decode("utf-8"),
                'path': parsed.path.decode("utf-8"),
                'params': parsed.params.decode("utf-8"),
                'query': parsed.query.decode("utf-8"),
                'fragment': parsed.fragment.decode("utf-8")
            }
        except Exception:
            logger.exception("Exception during referer parsing")
            return self.request.headers.get('Referer', None)

    def _get_ip(self):
        try:
            ip = self.request.headers.get("X-Forwarded-For", self.request.remote_ip)
            ip = ip.split(',')[-1].strip(ip)
            return ip
        except Exception:
            logger.exception(msg="Exception during ip parsing")
            return self.request.headers.get("X-Real-Ip")

    def _parse_user_agent(self):
        try:
            return str(user_agents.parse(
                self.request.headers.get('User-Agent')))
        except Exception as ex:
            logging.error(ex)
            return self.request.headers.get('User-Agent')

class SecurityManager(object):

    def __init__(self, app, secret=None, issuer=None):
        self._app = app
        self._secret = secret or app.settings.get('secret', None)
        self._issuer = issuer
        self._app.before_request_funcs.insert(0, self._before_request_hook)
        self._user_loader = None
        self._secret_loader = None

    def user_loader(self, f):
        self._user_loader = f
        return f

    def secret_loader(self, f):
        self._secret_loader = f
        return f

    async def _before_request_hook(self, handler, method_definition):
        required_permission = method_definition.get('secure', None)

        auth_header = handler.request.headers.get('Authorization', None)

        if not auth_header and not required_permission:
            handler.context.subject = Subject(request=handler.request)
            return

        if not auth_header and required_permission:
            raise AgentIsNotAuthorized(
                message="Authorization header is not found")

        if not auth_header.startswith('Bearer'):
            raise AgentIsNotAuthorized(
                message='Unsupported authorization scheme')

        jwt_token = auth_header.split(' ')[1]
        try:
            decoded = jwt.decode(
                jwt_token.strip(),
                key=self._secret,
                verify=True,
                algorithms="HS256",
                leeway=5
            )

            subject = Subject(decoded)
            if self._user_loader:
                logger.info("User loader has been found. Loading user...")
                subject.user = await self._user_loader(handler, method_definition, decoded)

            handler.context.subject = Subject(decoded, request=handler.request)

            if not required_permission:
                return

            user_permissions = decoded.get("permissions", None)
            for user_permission in user_permissions:
                if implies(user_permission, required_permission):
                    logger.info(
                        "User<{}> has permission<{}> to execute this action."
                        .format(decoded["prn"], required_permission))
                    return

            logger.info(
                    "User<username={}, permissions={}> doesn't has "
                    "required permission<{}> to execute this action"
                    .format(
                        decoded["prn"], user_permissions, required_permission))

            raise AgentIsNotAuthorized(message="User has no permission to use this API",
                                       required_permission=required_permission)


        except jwt.exceptions.InvalidTokenError as e:
            raise errors.FError(
                err_msg=str(e),
                err_code='errors.invalidTokenError',
                status_code=400,
                context={
                    'authorization_header': auth_header
                }
            )


_PART_DIVIDER = "."
_SUBPART_DIVIDER = ","
_WILDCARD_TOKEN = "*"


def partify(permission_string):
    """
    @type permission_string: str
    """
    if not permission_string:
        raise ValueError("Wildcard string cannot be none or empty")
    permission_string = permission_string.strip()

    _parts = []

    splitted_parts = permission_string.split(_PART_DIVIDER)
    for splitted_part in splitted_parts:
        subparts = splitted_part.lower().split(_SUBPART_DIVIDER)
        if not subparts:
            raise ValueError(
                    "Wildcard string cannot contains"
                    "parts with only dividers.")
        _parts.append(set(subparts))

    if not _parts:
        raise ValueError("Wildcard string cannot contain only dividers")

    return _parts



def implies(permission_1, permission_2):
    permission_parts = partify(permission_1)
    other_permission_parts = partify(permission_2)

    i = 0
    for other_permission_part in other_permission_parts:
        #: if this permission has less part than other permission,
        #: everything after the number of parts contained
        #: in this permission is implied.
        #: eg: com.admin implies com.admin.read
        if len(permission_parts) - 1 < i:
            return True
        elif _WILDCARD_TOKEN not in permission_parts[i] and \
                not permission_parts[i].issuperset(other_permission_part):
            return False
        i += 1

    for i in range(i, len(permission_parts)-1):
        if _WILDCARD_TOKEN not in permission_parts[i]:
            return False

    return True

def implies_any(permission_collection, permission):
    for _permission in permission_collection:
        if implies(_permission, permission):
            return True

    return False
